@@ -1,7 +1,26 @@
-var Buffer = require('buffer').Buffer;
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-var StringDecoder = exports.StringDecoder = function (encoding) {
-  this.encoding = (encoding || 'utf8').toLowerCase().replace(/[-_]/,'');
+var StringDecoder = exports.StringDecoder = function(encoding) {
+  this.encoding = (encoding || 'utf8').toLowerCase().replace(/[-_]/, '');
   if (this.encoding === 'utf8') {
     this.charBuffer = new Buffer(4);
     this.charReceived = 0;
@@ -10,7 +29,7 @@ var StringDecoder = exports.StringDecoder = function (encoding) {
 };
 
 
-StringDecoder.prototype.write = function (buffer) {
+StringDecoder.prototype.write = function(buffer) {
   // If not utf8...
   if (this.encoding !== 'utf8') {
     return buffer.toString(this.encoding);
@@ -20,9 +39,9 @@ StringDecoder.prototype.write = function (buffer) {
   // if our last write ended with an incomplete multibyte character
   if (this.charLength) {
     // determine how many remaining bytes this buffer has to offer for this char
-    var i = (buffer.length >= this.charLength - this.charReceived)
-      ? this.charLength - this.charReceived
-      : buffer.length;
+    var i = (buffer.length >= this.charLength - this.charReceived) ?
+                this.charLength - this.charReceived :
+                buffer.length;
 
     // add the new bytes to the char buffer
     buffer.copy(this.charBuffer, this.charReceived, 0, i);
@@ -48,9 +67,10 @@ StringDecoder.prototype.write = function (buffer) {
   // determine how many bytes we have to check at the end of this buffer
   var i = (buffer.length >= 3) ? 3 : buffer.length;
 
-  // figure out if one of the last i bytes of our buffer announces an incomplete char
+  // Figure out if one of the last i bytes of our buffer announces an
+  // incomplete char.
   for (; i > 0; i--) {
-    c = buffer[buffer.length - i];
+    var c = buffer[buffer.length - i];
 
     // See http://en.wikipedia.org/wiki/UTF-8#Description
 
