@@ -1,38 +1,87 @@
-#include <node_net.h>
-#include <v8.h>
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
+
 
 #include <node.h>
 #include <node_buffer.h>
+#include <node_net.h>
 
+#include <v8.h>
+
+#include <errno.h>
 #include <string.h>
 #include <stdlib.h>
 
 #include <sys/types.h>
-#include <sys/socket.h>
-#include <sys/un.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <arpa/inet.h> /* inet_pton */
 
-#include <netdb.h>
+#ifdef __MINGW32__
+# include <platform_win32.h>
+# include <platform_win32_winsock.h>
+#endif
 
-#include <netinet/in.h>
-#include <netinet/tcp.h>
-
-#include <sys/ioctl.h>
+#ifdef __POSIX__
+# include <sys/ioctl.h>
+# include <sys/socket.h>
+# include <sys/un.h>
+# include <arpa/inet.h> /* inet_pton */
+# include <netdb.h>
+# include <netinet/in.h>
+# include <netinet/tcp.h>
+#endif
 
 #ifdef __linux__
 # include <linux/sockios.h> /* For the SIOCINQ / FIONREAD ioctl */
 #endif
+
 /* Non-linux platforms like OS X define this ioctl elsewhere */
 #ifndef FIONREAD
-#include <sys/filio.h>
+# include <sys/filio.h>
 #endif
 
-#include <errno.h>
-
 #ifdef __OpenBSD__
-#include <sys/uio.h>
+# include <sys/uio.h>
+#endif
+
+/*
+ * HACK to use inet_pton/inet_ntop from c-ares because mingw32 doesn't have it
+ * This trick is used in node_ares.cc as well
+ * TODO fixme
+ */
+#ifdef __MINGW32__
+  extern "C" {
+#   include <inet_net_pton.h>
+#   include <inet_ntop.h>
+  }
+
+# define inet_pton ares_inet_pton
+# define inet_ntop ares_inet_ntop
+#endif
+
+// SHUT_* constants aren't available on windows but there are 1:1 equivalents
+#ifdef __MINGW32__
+# define SHUT_RD   SD_RECEIVE
+# define SHUT_WR   SD_SEND
+# define SHUT_RDWR SD_BOTH
 #endif
 
 #define ARRAY_SIZE(a) (sizeof(a) / sizeof(*(a)))
@@ -63,23 +112,41 @@ static Persistent<FunctionTemplate> recv_msg_template;
           String::New("Bad file descriptor argument"))); \
   }
 
+
 static inline bool SetCloseOnExec(int fd) {
+#ifdef __POSIX__
   return (fcntl(fd, F_SETFD, FD_CLOEXEC) != -1);
+#else // __MINGW32__
+  return SetHandleInformation(reinterpret_cast<HANDLE>(_get_osfhandle(fd)),
+                              HANDLE_FLAG_INHERIT, 0) != 0;
+#endif
 }
 
 
 static inline bool SetNonBlock(int fd) {
+#ifdef __MINGW32__
+  unsigned long value = 1;
+  return (ioctlsocket(_get_osfhandle(fd), FIONBIO, &value) == 0);
+#else // __POSIX__
   return (fcntl(fd, F_SETFL, O_NONBLOCK) != -1);
+#endif
 }
 
 
 static inline bool SetSockFlags(int fd) {
+#ifdef __MINGW32__
+  BOOL flags = TRUE;
+  setsockopt(_get_osfhandle(fd), SOL_SOCKET, SO_REUSEADDR, (const char *)&flags, sizeof(flags));
+#else // __POSIX__
   int flags = 1;
   setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *)&flags, sizeof(flags));
+#endif
   return SetNonBlock(fd) && SetCloseOnExec(fd);
 }
 
 
+#ifdef __POSIX__
+
 // Creates nonblocking pipe
 static Handle<Value> Pipe(const Arguments& args) {
   HandleScope scope;
@@ -125,6 +192,8 @@ static Handle<Value> SocketPair(const Arguments& args) {
   return scope.Close(a);
 }
 
+#endif
+
 
 // Creates a new non-blocking socket fd
 // t.socket("TCP");
@@ -182,7 +251,11 @@ static Handle<Value> Socket(const Arguments& args) {
     }
   }
 
+#ifdef __POSIX__
   int fd = socket(domain, type, 0);
+#else // __MINGW32__
+  int fd = _open_osfhandle(socket(domain, type, 0), 0);
+#endif
 
   if (fd < 0) return ThrowException(ErrnoException(errno, "socket"));
 
@@ -213,10 +286,12 @@ static socklen_t addrlen;
 static inline Handle<Value> ParseAddressArgs(Handle<Value> first,
                                              Handle<Value> second,
                                              bool is_bind) {
-  static struct sockaddr_un un;
   static struct sockaddr_in in;
   static struct sockaddr_in6 in6;
 
+#ifdef __POSIX__ // No unix sockets on windows
+  static struct sockaddr_un un;
+
   if (first->IsString() && !second->IsString()) {
     // UNIX
     String::Utf8Value path(first->ToString());
@@ -233,6 +308,11 @@ static inline Handle<Value> ParseAddressArgs(Handle<Value> first,
     addrlen = sizeof(un) - sizeof(un.sun_path) + path.length() + 1;
 
   } else {
+#else // __MINGW32__
+  if (first->IsString() && !second->IsString()) {
+    return ErrnoException(errno, "ParseAddressArgs", "Unix sockets are not supported on windows");
+  } else {
+#endif
     // TCP or UDP
     memset(&in, 0, sizeof in);
     memset(&in6, 0, sizeof in6);
@@ -284,14 +364,23 @@ static Handle<Value> Bind(const Arguments& args) {
   if (!error.IsEmpty()) return ThrowException(error);
 
   int flags = 1;
+
+#ifdef __POSIX__
   setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *)&flags, sizeof(flags));
 
-  int r = bind(fd, addr, addrlen);
-
-  if (r < 0) {
+  if (0 > bind(fd, addr, addrlen)) {
     return ThrowException(ErrnoException(errno, "bind"));
   }
 
+#else // __MINGW32__
+  SOCKET handle =  _get_osfhandle(fd);
+  setsockopt(handle, SOL_SOCKET, SO_REUSEADDR, (char *)&flags, sizeof(flags));
+
+  if (SOCKET_ERROR == bind(handle, addr, addrlen)) {
+    return ThrowException(ErrnoException(WSAGetLastError(), "bind"));
+  }
+#endif // __MINGW32__
+
   return Undefined();
 }
 
@@ -301,6 +390,7 @@ static Handle<Value> Close(const Arguments& args) {
 
   FD_ARG(args[0])
 
+  // Windows: this is not a winsock operation, don't use _get_osfhandle here!
   if (0 > close(fd)) {
     return ThrowException(ErrnoException(errno, "close"));
   }
@@ -334,9 +424,15 @@ static Handle<Value> Shutdown(const Arguments& args) {
     }
   }
 
+#ifdef __POSIX__
   if (0 > shutdown(fd, how)) {
     return ThrowException(ErrnoException(errno, "shutdown"));
   }
+#else // __MINGW32__
+  if (SOCKET_ERROR == shutdown(_get_osfhandle(fd), how)) {
+    return ThrowException(ErrnoException(WSAGetLastError(), "shutdown"));
+  }
+#endif // __MINGW32__
 
   return Undefined();
 }
@@ -363,15 +459,29 @@ static Handle<Value> Connect(const Arguments& args) {
   Handle<Value> error = ParseAddressArgs(args[1], args[2], false);
   if (!error.IsEmpty()) return ThrowException(error);
 
+#ifdef __POSIX__
   int r = connect(fd, addr, addrlen);
 
   if (r < 0 && errno != EINPROGRESS) {
     return ThrowException(ErrnoException(errno, "connect"));
   }
+#else // __MINGW32__
+  int r = connect(_get_osfhandle(fd), addr, addrlen);
+
+  if (r == SOCKET_ERROR) {
+    int wsaErrno = WSAGetLastError();
+    if (wsaErrno != WSAEWOULDBLOCK && wsaErrno != WSAEINPROGRESS) {
+      return ThrowException(ErrnoException(wsaErrno, "connect"));
+    }
+  }
+#endif // __MINGW32__
 
   return Undefined();
 }
 
+
+#ifdef __POSIX__
+
 #define ADDRESS_TO_JS(info, address_storage, addrlen) \
 do { \
   char ip[INET6_ADDRSTRLEN]; \
@@ -421,6 +531,40 @@ do { \
   } \
 } while (0)
 
+#else // __MINGW32__
+
+#define ADDRESS_TO_JS(info, address_storage, addrlen) \
+do { \
+  char ip[INET6_ADDRSTRLEN]; \
+  int port; \
+  struct sockaddr_in *a4; \
+  struct sockaddr_in6 *a6; \
+  if (addrlen == 0) { \
+    (info)->Set(address_symbol, String::Empty()); \
+  } else { \
+    switch ((address_storage).ss_family) { \
+      case AF_INET6: \
+        a6 = (struct sockaddr_in6*)&(address_storage); \
+        inet_ntop(AF_INET6, &(a6->sin6_addr), ip, INET6_ADDRSTRLEN); \
+        port = ntohs(a6->sin6_port); \
+        (info)->Set(address_symbol, String::New(ip)); \
+        (info)->Set(port_symbol, Integer::New(port)); \
+        break; \
+      case AF_INET: \
+        a4 = (struct sockaddr_in*)&(address_storage); \
+        inet_ntop(AF_INET, &(a4->sin_addr), ip, INET6_ADDRSTRLEN); \
+        port = ntohs(a4->sin_port); \
+        (info)->Set(address_symbol, String::New(ip)); \
+        (info)->Set(port_symbol, Integer::New(port)); \
+        break; \
+      default: \
+        (info)->Set(address_symbol, String::Empty()); \
+    } \
+  } \
+} while (0)
+
+#endif // __MINGW32__
+
 
 static Handle<Value> GetSockName(const Arguments& args) {
   HandleScope scope;
@@ -430,16 +574,20 @@ static Handle<Value> GetSockName(const Arguments& args) {
   struct sockaddr_storage address_storage;
   socklen_t len = sizeof(struct sockaddr_storage);
 
-  int r = getsockname(fd, (struct sockaddr *) &address_storage, &len);
-
-  if (r < 0) {
+#ifdef __POSIX__
+  if (0 > getsockname(fd, (struct sockaddr *) &address_storage, &len)) {
     return ThrowException(ErrnoException(errno, "getsockname"));
   }
 
+#else // __MINGW32__
+  if (SOCKET_ERROR == getsockname(_get_osfhandle(fd),
+      (struct sockaddr *) &address_storage, &len)) {
+    return ThrowException(ErrnoException(WSAGetLastError(), "getsockname"));
+  }
+#endif // __MINGW32__
+
   Local<Object> info = Object::New();
-
   ADDRESS_TO_JS(info, address_storage, len);
-
   return scope.Close(info);
 }
 
@@ -452,16 +600,20 @@ static Handle<Value> GetPeerName(const Arguments& args) {
   struct sockaddr_storage address_storage;
   socklen_t len = sizeof(struct sockaddr_storage);
 
-  int r = getpeername(fd, (struct sockaddr *) &address_storage, &len);
-
-  if (r < 0) {
-    return ThrowException(ErrnoException(errno, "getsockname"));
+#ifdef __POSIX__
+  if (0 > getpeername(fd, (struct sockaddr *) &address_storage, &len)) {
+    return ThrowException(ErrnoException(errno, "getpeername"));
   }
 
+#else // __MINGW32__
+  if (SOCKET_ERROR == getpeername(_get_osfhandle(fd),
+      (struct sockaddr *) &address_storage, &len)) {
+    return ThrowException(ErrnoException(WSAGetLastError(), "getpeername"));
+  }
+#endif // __MINGW32__
+
   Local<Object> info = Object::New();
-
   ADDRESS_TO_JS(info, address_storage, len);
-
   return scope.Close(info);
 }
 
@@ -472,10 +624,15 @@ static Handle<Value> Listen(const Arguments& args) {
   FD_ARG(args[0])
   int backlog = args[1]->IsInt32() ? args[1]->Int32Value() : 128;
 
+#ifdef __POSIX__
   if (0 > listen(fd, backlog)) {
     return ThrowException(ErrnoException(errno, "listen"));
   }
-
+#else // __MINGW32__
+  if (SOCKET_ERROR == listen(_get_osfhandle(fd), backlog)) {
+    return ThrowException(ErrnoException(WSAGetLastError(), "listen"));
+  }
+#endif
 
   return Undefined();
 }
@@ -498,6 +655,7 @@ static Handle<Value> Accept(const Arguments& args) {
   struct sockaddr_storage address_storage;
   socklen_t len = sizeof(struct sockaddr_storage);
 
+#ifdef __POSIX__
   int peer_fd = accept(fd, (struct sockaddr*) &address_storage, &len);
 
   if (peer_fd < 0) {
@@ -505,9 +663,24 @@ static Handle<Value> Accept(const Arguments& args) {
     if (errno == ECONNABORTED) return scope.Close(Null());
     return ThrowException(ErrnoException(errno, "accept"));
   }
+#else // __MINGW32__
+  SOCKET peer_handle = accept(_get_osfhandle(fd), (struct sockaddr*) &address_storage, &len);
+
+  if (peer_handle == INVALID_SOCKET) {
+    int wsaErrno = WSAGetLastError();
+    if (wsaErrno == WSAEWOULDBLOCK) return scope.Close(Null());
+    return ThrowException(ErrnoException(wsaErrno, "accept"));
+  }
+
+  int peer_fd = _open_osfhandle(peer_handle, 0);
+#endif // __MINGW32__
 
   if (!SetSockFlags(peer_fd)) {
+#ifdef __POSIX__
     int fcntl_errno = errno;
+#else // __MINGW32__
+    int fcntl_errno = WSAGetLastError();
+#endif // __MINGW32__
     close(peer_fd);
     return ThrowException(ErrnoException(fcntl_errno, "fcntl"));
   }
@@ -529,11 +702,20 @@ static Handle<Value> SocketError(const Arguments& args) {
 
   int error;
   socklen_t len = sizeof(int);
+
+#ifdef __POSIX__
   int r = getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len);
 
   if (r < 0) {
     return ThrowException(ErrnoException(errno, "getsockopt"));
   }
+#else // __MINGW32__
+  int r = getsockopt(_get_osfhandle(fd), SOL_SOCKET, SO_ERROR, (char*)&error, &len);
+
+  if (r < 0) {
+    return ThrowException(ErrnoException(WSAGetLastError(), "getsockopt"));
+  }
+#endif
 
   return scope.Close(Integer::New(error));
 }
@@ -557,30 +739,45 @@ static Handle<Value> Read(const Arguments& args) {
           String::New("Second argument should be a buffer")));
   }
 
-  Buffer * buffer = ObjectWrap::Unwrap<Buffer>(args[1]->ToObject());
+  Local<Object> buffer_obj = args[1]->ToObject();
+  char *buffer_data = Buffer::Data(buffer_obj);
+  size_t buffer_length = Buffer::Length(buffer_obj);
 
   size_t off = args[2]->Int32Value();
-  if (off >= buffer->length()) {
+  if (off >= buffer_length) {
     return ThrowException(Exception::Error(
           String::New("Offset is out of bounds")));
   }
 
   size_t len = args[3]->Int32Value();
-  if (off + len > buffer->length()) {
+  if (off + len > buffer_length) {
     return ThrowException(Exception::Error(
           String::New("Length is extends beyond buffer")));
   }
 
-  ssize_t bytes_read = read(fd, (char*)buffer->data() + off, len);
+#ifdef __POSIX__
+  ssize_t bytes_read = read(fd, (char*)buffer_data + off, len);
 
   if (bytes_read < 0) {
     if (errno == EAGAIN || errno == EINTR) return Null();
     return ThrowException(ErrnoException(errno, "read"));
   }
+#else // __MINGW32__
+   // read() doesn't work for overlapped sockets (the only usable
+   // type of sockets) so recv() is used here.
+  ssize_t bytes_read = recv(_get_osfhandle(fd), (char*)buffer_data + off, len, 0);
+
+  if (bytes_read < 0) {
+    int wsaErrno = WSAGetLastError();
+    if (wsaErrno == WSAEWOULDBLOCK || wsaErrno == WSAEINTR) return Null();
+    return ThrowException(ErrnoException(wsaErrno, "read"));
+  }
+#endif
 
   return scope.Close(Integer::New(bytes_read));
 }
 
+
 //  var info = t.recvfrom(fd, buffer, offset, length, flags);
 //    info.size // bytes read
 //    info.port // from port
@@ -602,16 +799,18 @@ static Handle<Value> RecvFrom(const Arguments& args) {
           String::New("Second argument should be a buffer")));
   }
 
-  Buffer * buffer = ObjectWrap::Unwrap<Buffer>(args[1]->ToObject());
+  Local<Object> buffer_obj = args[1]->ToObject();
+  char *buffer_data = Buffer::Data(buffer_obj);
+  size_t buffer_length = Buffer::Length(buffer_obj);
 
   size_t off = args[2]->Int32Value();
-  if (off >= buffer->length()) {
+  if (off >= buffer_length) {
     return ThrowException(Exception::Error(
           String::New("Offset is out of bounds")));
   }
 
   size_t len = args[3]->Int32Value();
-  if (off + len > buffer->length()) {
+  if (off + len > buffer_length) {
     return ThrowException(Exception::Error(
           String::New("Length is extends beyond buffer")));
   }
@@ -621,7 +820,8 @@ static Handle<Value> RecvFrom(const Arguments& args) {
   struct sockaddr_storage address_storage;
   socklen_t addrlen = sizeof(struct sockaddr_storage);
 
-  ssize_t bytes_read = recvfrom(fd, (char*)buffer->data() + off, len, flags,
+#ifdef __POSIX__
+  ssize_t bytes_read = recvfrom(fd, (char*)buffer_data + off, len, flags,
                                 (struct sockaddr*) &address_storage, &addrlen);
 
   if (bytes_read < 0) {
@@ -629,6 +829,17 @@ static Handle<Value> RecvFrom(const Arguments& args) {
     return ThrowException(ErrnoException(errno, "read"));
   }
 
+#else // __MINGW32__
+  ssize_t bytes_read = recvfrom(_get_osfhandle(fd), (char*)buffer_data + off,
+      len, flags, (struct sockaddr*) &address_storage, &addrlen);
+
+  if (bytes_read == SOCKET_ERROR) {
+    int wsaErrno = WSAGetLastError();
+    if (wsaErrno == WSAEWOULDBLOCK || wsaErrno == WSAEINTR) return Null();
+    return ThrowException(ErrnoException(wsaErrno, "read"));
+  }
+#endif
+
   Local<Object> info = Object::New();
 
   info->Set(size_symbol, Integer::New(bytes_read));
@@ -639,6 +850,8 @@ static Handle<Value> RecvFrom(const Arguments& args) {
 }
 
 
+#ifdef __POSIX__
+
 // bytesRead = t.recvMsg(fd, buffer, offset, length)
 // if (recvMsg.fd) {
 //   receivedFd = recvMsg.fd;
@@ -658,22 +871,24 @@ static Handle<Value> RecvMsg(const Arguments& args) {
           String::New("Second argument should be a buffer")));
   }
 
-  Buffer * buffer = ObjectWrap::Unwrap<Buffer>(args[1]->ToObject());
+  Local<Object> buffer_obj = args[1]->ToObject();
+  char *buffer_data = Buffer::Data(buffer_obj);
+  size_t buffer_length = Buffer::Length(buffer_obj);
 
   size_t off = args[2]->Int32Value();
-  if (off >= buffer->length()) {
+  if (off >= buffer_length) {
     return ThrowException(Exception::Error(
           String::New("Offset is out of bounds")));
   }
 
   size_t len = args[3]->Int32Value();
-  if (off + len > buffer->length()) {
+  if (off + len > buffer_length) {
     return ThrowException(Exception::Error(
           String::New("Length is extends beyond buffer")));
   }
 
   struct iovec iov[1];
-  iov[0].iov_base = (char*)buffer->data() + off;
+  iov[0].iov_base = (char*)buffer_data + off;
   iov[0].iov_len = len;
 
   struct msghdr msg;
@@ -733,6 +948,8 @@ static Handle<Value> RecvMsg(const Arguments& args) {
   return scope.Close(Integer::New(bytes_read));
 }
 
+#endif // __POSIX__
+
 
 //  var bytesWritten = t.write(fd, buffer, offset, length);
 //  returns null on EAGAIN or EINTR, raises an exception on all other errors
@@ -751,21 +968,24 @@ static Handle<Value> Write(const Arguments& args) {
           String::New("Second argument should be a buffer")));
   }
 
-  Buffer * buffer = ObjectWrap::Unwrap<Buffer>(args[1]->ToObject());
+  Local<Object> buffer_obj = args[1]->ToObject();
+  char *buffer_data = Buffer::Data(buffer_obj);
+  size_t buffer_length = Buffer::Length(buffer_obj);
 
   size_t off = args[2]->Int32Value();
-  if (off >= buffer->length()) {
+  if (off >= buffer_length) {
     return ThrowException(Exception::Error(
           String::New("Offset is out of bounds")));
   }
 
   size_t len = args[3]->Int32Value();
-  if (off + len > buffer->length()) {
+  if (off + len > buffer_length) {
     return ThrowException(Exception::Error(
           String::New("Length is extends beyond buffer")));
   }
 
-  ssize_t written = write(fd, (char*)buffer->data() + off, len);
+#ifdef __POSIX__
+  ssize_t written = write(fd, buffer_data + off, len);
 
   if (written < 0) {
     if (errno == EAGAIN || errno == EINTR) {
@@ -773,11 +993,26 @@ static Handle<Value> Write(const Arguments& args) {
     }
     return ThrowException(ErrnoException(errno, "write"));
   }
+#else // __MINGW32__
+  // write() doesn't work for overlapped sockets (the only usable
+  // type of sockets) so send() is used.
+  ssize_t written = send(_get_osfhandle(fd), buffer_data + off, len, 0);
+
+  if (written < 0) {
+    int wsaErrno = WSAGetLastError();
+    if (wsaErrno == WSAEWOULDBLOCK || wsaErrno == WSAEINTR) {
+      return scope.Close(Integer::New(0));
+    }
+    return ThrowException(ErrnoException(wsaErrno, "write"));
+  }
+#endif // __MINGW32__
 
   return scope.Close(Integer::New(written));
 }
 
 
+#ifdef __POSIX__
+
 // var bytes = sendmsg(fd, buf, off, len, fd, flags);
 //
 // Write a buffer with optional offset and length to the given file
@@ -809,7 +1044,9 @@ static Handle<Value> SendMsg(const Arguments& args) {
       String::New("Expected either a string or a buffer")));
   }
 
-  Buffer *buf = ObjectWrap::Unwrap<Buffer>(args[1]->ToObject());
+  Local<Object> buffer_obj = args[1]->ToObject();
+  char *buffer_data = Buffer::Data(buffer_obj);
+  size_t buffer_length = Buffer::Length(buffer_obj);
 
   size_t offset = 0;
   if (args.Length() >= 3 && !args[2]->IsUndefined()) {
@@ -819,13 +1056,13 @@ static Handle<Value> SendMsg(const Arguments& args) {
     }
 
     offset = args[2]->Uint32Value();
-    if (offset >= buf->length()) {
+    if (offset >= buffer_length) {
       return ThrowException(Exception::Error(
         String::New("Offset into buffer too large")));
     }
   }
 
-  size_t length = buf->length() - offset;
+  size_t length = buffer_length - offset;
   if (args.Length() >= 4 && !args[3]->IsUndefined()) {
     if (!args[3]->IsUint32()) {
       return ThrowException(Exception::TypeError(
@@ -833,13 +1070,13 @@ static Handle<Value> SendMsg(const Arguments& args) {
     }
 
     length = args[3]->Uint32Value();
-    if (offset + length > buf->length()) {
+    if (offset + length > buffer_length) {
       return ThrowException(Exception::Error(
         String::New("offset + length beyond buffer length")));
     }
   }
 
-  iov.iov_base = buf->data() + offset;
+  iov.iov_base = buffer_data + offset;
   iov.iov_len = length;
 
   int fd_to_send = -1;
@@ -899,6 +1136,9 @@ static Handle<Value> SendMsg(const Arguments& args) {
   return scope.Close(Integer::New(written));
 }
 
+#endif // __POSIX__
+
+
 // var bytes = sendto(fd, buf, off, len, flags, destination port, desitnation address);
 //
 // Write a buffer with optional offset and length to the given file
@@ -930,7 +1170,9 @@ static Handle<Value> SendTo(const Arguments& args) {
       String::New("Expected either a string or a buffer")));
   }
 
-  Buffer *buf = ObjectWrap::Unwrap<Buffer>(args[1]->ToObject());
+  Local<Object> buffer_obj = args[1]->ToObject();
+  char *buffer_data = Buffer::Data(buffer_obj);
+  size_t buffer_length = Buffer::Length(buffer_obj);
 
   size_t offset = 0;
   if (args.Length() >= 3 && !args[2]->IsUndefined()) {
@@ -940,13 +1182,13 @@ static Handle<Value> SendTo(const Arguments& args) {
     }
 
     offset = args[2]->Uint32Value();
-    if (offset >= buf->length()) {
+    if (offset >= buffer_length) {
       return ThrowException(Exception::Error(
         String::New("Offset into buffer too large")));
     }
   }
 
-  size_t length = buf->length() - offset;
+  size_t length = buffer_length - offset;
   if (args.Length() >= 4 && !args[3]->IsUndefined()) {
     if (!args[3]->IsUint32()) {
       return ThrowException(Exception::TypeError(
@@ -954,7 +1196,7 @@ static Handle<Value> SendTo(const Arguments& args) {
     }
 
     length = args[3]->Uint32Value();
-    if (offset + length > buf->length()) {
+    if (offset + length > buffer_length) {
       return ThrowException(Exception::Error(
         String::New("offset + length beyond buffer length")));
     }
@@ -973,13 +1215,26 @@ static Handle<Value> SendTo(const Arguments& args) {
   Handle<Value> error = ParseAddressArgs(args[5], args[6], false);
   if (!error.IsEmpty()) return ThrowException(error);
 
-  ssize_t written = sendto(fd, buf->data() + offset, length, flags, addr, addrlen);
+#ifdef __POSIX__
+  ssize_t written = sendto(fd, buffer_data + offset, length, flags, addr,
+      addrlen);
 
   if (written < 0) {
     if (errno == EAGAIN || errno == EINTR) return Null();
-    return ThrowException(ErrnoException(errno, "sendmsg"));
+    return ThrowException(ErrnoException(errno, "sendto"));
   }
 
+#else // __MINGW32__
+  ssize_t written = sendto(_get_osfhandle(fd), buffer_data + offset, length,
+      flags, addr, addrlen);
+
+  if (written == SOCKET_ERROR) {
+    int wsaErrno = WSAGetLastError();
+    if (wsaErrno == WSAEWOULDBLOCK || wsaErrno == WSAEINTR) return Null();
+    return ThrowException(ErrnoException(wsaErrno, "sendto"));
+  }
+#endif // __MINGW32__
+
   /* Note that the FD isn't explicitly closed here, this
    * happens in the JS */
 
@@ -994,32 +1249,51 @@ static Handle<Value> ToRead(const Arguments& args) {
 
   FD_ARG(args[0])
 
+#ifdef __POSIX__
   int value;
-  int r = ioctl(fd, FIONREAD, &value);
 
-  if (r < 0) {
+  if (0 > ioctl(fd, FIONREAD, &value)) {
     return ThrowException(ErrnoException(errno, "ioctl"));
   }
 
+#else // __MINGW32__
+  unsigned long value;
+
+  if (SOCKET_ERROR == ioctlsocket(_get_osfhandle(fd), FIONREAD, &value)) {
+    return ThrowException(ErrnoException(WSAGetLastError(), "ioctlsocket"));
+  }
+#endif // __MINGW32__
+
   return scope.Close(Integer::New(value));
 }
 
 
 static Handle<Value> SetNoDelay(const Arguments& args) {
-  int flags, r;
   HandleScope scope;
 
   FD_ARG(args[0])
 
-  flags = args[1]->IsFalse() ? 0 : 1;
-  r = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (void *)&flags, sizeof(flags));
+#ifdef __POSIX__
+  int flags = args[1]->IsFalse() ? 0 : 1;
 
-  if (r < 0) {
+  if (0 > setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (void *)&flags,
+      sizeof(flags))) {
     return ThrowException(ErrnoException(errno, "setsockopt"));
   }
+
+#else // __MINGW32__
+  BOOL flags = args[1]->IsFalse() ? FALSE : TRUE;
+
+  if (SOCKET_ERROR == setsockopt(_get_osfhandle(fd), IPPROTO_TCP, TCP_NODELAY,
+      (const char *)&flags, sizeof(flags))) {
+    return ThrowException(ErrnoException(WSAGetLastError(), "setsockopt"));
+  }
+#endif // __MINGW32__
+
   return Undefined();
 }
 
+
 static Handle<Value> SetKeepAlive(const Arguments& args) {
   int r;
   HandleScope scope;
@@ -1034,7 +1308,9 @@ static Handle<Value> SetKeepAlive(const Arguments& args) {
     time = args[2]->Int32Value();
   }
 
+#ifdef __POSIX__
   int flags = enable ? 1 : 0;
+
   r = setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (void *)&flags, sizeof(flags));
   if ((time > 0)&&(r >= 0)) {
 #if defined(__APPLE__)
@@ -1048,28 +1324,58 @@ static Handle<Value> SetKeepAlive(const Arguments& args) {
   if (r < 0) {
     return ThrowException(ErrnoException(errno, "setsockopt"));
   }
+
+#else // __MINGW32__
+  SOCKET handle = (SOCKET)_get_osfhandle(fd);
+  BOOL flags = enable ? TRUE : FALSE;
+
+  r = setsockopt(handle, SOL_SOCKET, SO_KEEPALIVE, (const char *)&flags,
+      sizeof(flags));
+  // Could set the timeout here, using WSAIoctl(SIO_KEEPALIVE_VALS),
+  // but ryah thinks it is not necessary, and mingw is missing mstcpip.h anyway.
+  if (r == SOCKET_ERROR) {
+    return ThrowException(ErrnoException(WSAGetLastError(), "setsockopt"));
+  }
+#endif
+
   return Undefined();
 }
 
+
 static Handle<Value> SetBroadcast(const Arguments& args) {
-  int flags, r;
   HandleScope scope;
 
   FD_ARG(args[0])
 
-  flags = args[1]->IsFalse() ? 0 : 1;
-  r = setsockopt(fd, SOL_SOCKET, SO_BROADCAST, (void *)&flags, sizeof(flags));
+#ifdef __POSIX__
+  int flags = args[1]->IsFalse() ? 0 : 1;
 
-  if (r < 0) {
+  if (0 > setsockopt(fd, SOL_SOCKET, SO_BROADCAST, (void *)&flags,
+      sizeof(flags))) {
     return ThrowException(ErrnoException(errno, "setsockopt"));
-  } else {
-    return scope.Close(Integer::New(flags));
   }
+
+#else // __MINGW32__
+  BOOL flags = args[1]->IsFalse() ? FALSE : TRUE;
+
+  if (SOCKET_ERROR == setsockopt(_get_osfhandle(fd), SOL_SOCKET, SO_BROADCAST,
+      (const char *)&flags, sizeof(flags))) {
+    return ThrowException(ErrnoException(WSAGetLastError(), "setsockopt"));
+  }
+#endif
+
+  return Undefined();
 }
 
 static Handle<Value> SetTTL(const Arguments& args) {
   HandleScope scope;
 
+#ifdef __POSIX__
+  int newttl;
+#else // __MINGW32__
+  DWORD newttl;
+#endif
+
   if (args.Length() != 2) {
     return ThrowException(Exception::TypeError(
       String::New("Takes exactly two arguments: fd, new TTL")));
@@ -1077,18 +1383,65 @@ static Handle<Value> SetTTL(const Arguments& args) {
 
   FD_ARG(args[0]);
 
-  if (! args[1]->IsInt32()) {
+  if (!args[1]->IsInt32()) {
     return ThrowException(Exception::TypeError(
       String::New("Argument must be a number")));
   }
-  
-  int newttl = args[1]->Int32Value();
+
+  newttl = args[1]->Int32Value();
+
   if (newttl < 1 || newttl > 255) {
     return ThrowException(Exception::TypeError(
       String::New("new TTL must be between 1 and 255")));
   }
 
-  int r = setsockopt(fd, IPPROTO_IP, IP_TTL, (void *)&newttl, sizeof(newttl));
+#ifdef __POSIX__
+  int r = setsockopt(fd,
+                     IPPROTO_IP,
+                     IP_TTL,
+                     reinterpret_cast<void*>(&newttl),
+                     sizeof(newttl));
+  if (r < 0) {
+    return ThrowException(ErrnoException(errno, "setsockopt"));
+  }
+
+#else // __MINGW32__
+  if (SOCKET_ERROR > setsockopt(_get_osfhandle(fd), IPPROTO_IP, IP_TTL,
+      (const char *)&newttl, sizeof(newttl))) {
+    return ThrowException(ErrnoException(WSAGetLastError(), "setsockopt"));
+  }
+
+#endif // __MINGW32__
+
+  return scope.Close(Integer::New(newttl));
+}
+
+
+#ifdef  __POSIX__
+
+static Handle<Value> SetMulticastTTL(const Arguments& args) {
+  HandleScope scope;
+
+  if (args.Length() != 2) {
+    return ThrowException(Exception::TypeError(
+      String::New("Takes exactly two arguments: fd, new MulticastTTL")));
+  }
+
+  FD_ARG(args[0]);
+
+  if (!args[1]->IsInt32()) {
+    return ThrowException(Exception::TypeError(
+      String::New("Argument must be a number")));
+  }
+
+  int newttl = args[1]->Int32Value();
+  if (newttl < 0 || newttl > 255) {
+    return ThrowException(Exception::TypeError(
+      String::New("new MulticastTTL must be between 0 and 255")));
+  }
+
+  int r = setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL,
+    reinterpret_cast<void*>(&newttl), sizeof(newttl));
 
   if (r < 0) {
     return ThrowException(ErrnoException(errno, "setsockopt"));
@@ -1097,6 +1450,74 @@ static Handle<Value> SetTTL(const Arguments& args) {
   }
 }
 
+static Handle<Value> SetMulticastLoopback(const Arguments& args) {
+  int flags, r;
+  HandleScope scope;
+
+  FD_ARG(args[0])
+
+  flags = args[1]->IsFalse() ? 0 : 1;
+  r = setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP,
+    reinterpret_cast<void*>(&flags), sizeof(flags));
+
+  if (r < 0) {
+    return ThrowException(ErrnoException(errno, "setsockopt"));
+  } else {
+    return scope.Close(Integer::New(flags));
+  }
+}
+
+static Handle<Value> SetMembership(const Arguments& args, int socketOption) {
+  HandleScope scope;
+
+  if (args.Length() < 2 || args.Length() > 3) {
+    return ThrowException(Exception::TypeError(
+      String::New("Takes arguments: fd, multicast group, multicast address")));
+  }
+
+  FD_ARG(args[0]);
+
+  struct ip_mreq mreq;
+  memset(&mreq, 0, sizeof(mreq));
+
+  // Multicast address (arg[1])
+  String::Utf8Value multicast_address(args[1]->ToString());
+  if (inet_pton(
+      AF_INET, *multicast_address, &(mreq.imr_multiaddr.s_addr)) <= 0) {
+    return ErrnoException(errno, "inet_pton", "Invalid multicast address");
+  }
+
+  // Interface address (arg[2] - optional, default:INADDR_ANY)
+  if (args.Length() < 3 || !args[2]->IsString()) {
+    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
+  } else {
+    String::Utf8Value multicast_interface(args[2]->ToString());
+    if (inet_pton(
+        AF_INET, *multicast_interface, &(mreq.imr_interface.s_addr)) <= 0) {
+      return ErrnoException(errno, "inet_pton", "Invalid multicast interface");
+    }
+  }
+
+  int r = setsockopt(fd, IPPROTO_IP, socketOption,
+    reinterpret_cast<void*>(&mreq), sizeof(mreq));
+
+  if (r < 0) {
+    return ThrowException(ErrnoException(errno, "setsockopt"));
+  } else {
+    return Undefined();
+  }
+}
+
+static Handle<Value> AddMembership(const Arguments& args) {
+  return SetMembership(args, IP_ADD_MEMBERSHIP);
+}
+
+static Handle<Value> DropMembership(const Arguments& args) {
+  return SetMembership(args, IP_DROP_MEMBERSHIP);
+}
+
+#endif // __POSIX__
+
 
 //
 // G E T A D D R I N F O
@@ -1298,20 +1719,25 @@ void InitNet(Handle<Object> target) {
 
   NODE_SET_METHOD(target, "write", Write);
   NODE_SET_METHOD(target, "read", Read);
-
-  NODE_SET_METHOD(target, "sendMsg", SendMsg);
-  NODE_SET_METHOD(target, "recvfrom", RecvFrom);
   NODE_SET_METHOD(target, "sendto", SendTo);
+  NODE_SET_METHOD(target, "recvfrom", RecvFrom);
+
+#ifdef __POSIX__
+  NODE_SET_METHOD(target, "sendMsg", SendMsg);
 
   recv_msg_template =
       Persistent<FunctionTemplate>::New(FunctionTemplate::New(RecvMsg));
   target->Set(String::NewSymbol("recvMsg"), recv_msg_template->GetFunction());
+#endif //__POSIX__
 
   NODE_SET_METHOD(target, "socket", Socket);
   NODE_SET_METHOD(target, "close", Close);
   NODE_SET_METHOD(target, "shutdown", Shutdown);
+
+#ifdef __POSIX__
   NODE_SET_METHOD(target, "pipe", Pipe);
   NODE_SET_METHOD(target, "socketpair", SocketPair);
+#endif // __POSIX__
 
   NODE_SET_METHOD(target, "connect", Connect);
   NODE_SET_METHOD(target, "bind", Bind);
@@ -1323,21 +1749,18 @@ void InitNet(Handle<Object> target) {
   NODE_SET_METHOD(target, "setBroadcast", SetBroadcast);
   NODE_SET_METHOD(target, "setTTL", SetTTL);
   NODE_SET_METHOD(target, "setKeepAlive", SetKeepAlive);
+#ifdef __POSIX__
+  NODE_SET_METHOD(target, "setMulticastTTL", SetMulticastTTL);
+  NODE_SET_METHOD(target, "setMulticastLoopback", SetMulticastLoopback);
+  NODE_SET_METHOD(target, "addMembership", AddMembership);
+  NODE_SET_METHOD(target, "dropMembership", DropMembership);
+#endif // __POSIX__
   NODE_SET_METHOD(target, "getsockname", GetSockName);
   NODE_SET_METHOD(target, "getpeername", GetPeerName);
   NODE_SET_METHOD(target, "getaddrinfo", GetAddrInfo);
   NODE_SET_METHOD(target, "isIP", IsIP);
   NODE_SET_METHOD(target, "errnoException", CreateErrnoException);
 
-  target->Set(String::NewSymbol("ENOENT"), Integer::New(ENOENT));
-  target->Set(String::NewSymbol("EMFILE"), Integer::New(EMFILE));
-  target->Set(String::NewSymbol("EINPROGRESS"), Integer::New(EINPROGRESS));
-  target->Set(String::NewSymbol("EINTR"), Integer::New(EINTR));
-  target->Set(String::NewSymbol("EACCES"), Integer::New(EACCES));
-  target->Set(String::NewSymbol("EPERM"), Integer::New(EPERM));
-  target->Set(String::NewSymbol("EADDRINUSE"), Integer::New(EADDRINUSE));
-  target->Set(String::NewSymbol("ECONNREFUSED"), Integer::New(ECONNREFUSED));
-
   errno_symbol          = NODE_PSYMBOL("errno");
   syscall_symbol        = NODE_PSYMBOL("syscall");
   fd_symbol             = NODE_PSYMBOL("fd");
